/* sha1.c
**
** Implementation of the SHA1 message digest algorithm.
**
*/
#include <stdio.h>
#include <string.h>

#include "bittypes.h"
#include "bcv.h"

#include "sha1.h"

#define u32_t vx_unsigned32_t

DEFCVF(b2u32i, u32i2b, vx_unsigned32_t)

#define K0 ((u32_t) 0x5A827999)
#define K1 ((u32_t) 0x6ED9EBA1)
#define K2 ((u32_t) 0x8F1BBCDC)
#define K3 ((u32_t) 0xCA62C1D6)

#define f0(B, C, D) (((B) & (C)) | ((~(B)) & (D)))
#define f1(B, C, D) ((B) ^ (C) ^ (D))
#define f2(B, C, D) (((B) & (C)) | ((B) & (D)) | ((C) & (D)))
#define f3(B, C, D) ((B) ^ (C) ^ (D))

#define lrot(n, X) (((X) << (n)) | ((X) >> (32 - (n))))

static void SHA1_trans (SHA1_CTX *ctx)
{
    int t;
    u32_t tmp, A, B, C, D, E, W[80];
    unsigned char *p = ctx->buf;
    for (t = 0; t < 16; ++t) {
	W[t] = b2u32i ((char *) p); p += sizeof(u32_t);
    }
    for (t = 16; t < 80; ++t) {
	tmp = W[t - 3] ^ W[t - 8] ^ W[t - 14] ^ W[t - 16];
	W[t] = lrot (1, tmp);
    }
    A = ctx->H[0];
    B = ctx->H[1];
    C = ctx->H[2];
    D = ctx->H[3];
    E = ctx->H[4];
    for (t = 0; t < 20; ++t) {
	tmp = lrot(5, A) + f0(B, C, D) + E + W[t] + K0;
	E = D; D = C; C = lrot(30, B); B = A; A = tmp;
    }
    for (t = 20; t < 40; ++t) {
	tmp = lrot(5, A) + f1(B, C, D) + E + W[t] + K1;
	E = D; D = C; C = lrot(30, B); B = A; A = tmp;
    }
    for (t = 40; t < 60; ++t) {
	tmp = lrot(5, A) + f2(B, C, D) + E + W[t] + K2;
	E = D; D = C; C = lrot(30, B); B = A; A = tmp;
    }
    for (t = 60; t < 80; ++t) {
	tmp = lrot(5, A) + f3(B, C, D) + E + W[t] + K3;
	E = D; D = C; C = lrot(30, B); B = A; A = tmp;
    }
    ctx->H[0] += A;
    ctx->H[1] += B;
    ctx->H[2] += C;
    ctx->H[3] += D;
    ctx->H[4] += E;
}

/* Hier kommen die eigentlich exportierten Funktionen */
void SHA1Update (SHA1_CTX *ctx, unsigned char *msg, size_t msglen)
{
    size_t sz;
    u32_t b_low = ctx->bits_low + ((u32_t) msglen << 3);
    if (b_low < ctx->bits_low) { ++ctx->bits_high; }
    ctx->bits_low = b_low;
    ctx->bits_high += (u32_t) msglen >> 29;
    sz = ctx->len;
    if (sz > 0) {
	if (sz + msglen < 64) {
	    memcpy (&ctx->buf[sz], msg, msglen);
	    ctx->len += msglen;
	    return;
	}
	memcpy (&ctx->buf[sz], msg, 64 - sz);
	msg += (64 - sz); msglen -= (64 - sz);
	SHA1_trans (ctx);
	ctx->len = 0;
    }
    while (msglen >= 64) {
	memcpy (ctx->buf, msg, 64);
	SHA1_trans (ctx);
	msg += 64; msglen -= 64;
    }
    ctx->len = msglen;
    if (msglen > 0) { memcpy (ctx->buf, msg, msglen); }
}

void SHA1Init (SHA1_CTX *ctx)
{
    ctx->H[0] = 0x67452301; ctx->H[1] = 0xEFCDAB89;
    ctx->H[2] = 0x98BADCFE; ctx->H[3] = 0x10325476;
    ctx->H[4] = 0xC3D2E1F0;
    memset (ctx->buf, '\0', 80);
    ctx->len = 0;
    ctx->bits_high = ctx->bits_low = (u32_t) 0;
}

void SHA1Final (unsigned char buf[20], SHA1_CTX *ctx)
{
    size_t ix;
    unsigned char *p;
    ctx->buf[ctx->len++] = (unsigned char) 0x80;
    memset (&ctx->buf[ctx->len], '\0', 64 - ctx->len);
    if (ctx->len > 56) {
	SHA1_trans (ctx);
	memset (ctx->buf, '\0', 64);
	ctx->len = 0;
    }
    p = &ctx->buf[56]; u32i2b (ctx->bits_high, (char *) p);
    p += sizeof(u32_t); u32i2b (ctx->bits_low, (char *) p);
    ctx->len = 64;
    SHA1_trans (ctx);
    for (ix = 0, p = buf; ix < 5; ++ix, p += sizeof(u32_t)) {
	u32i2b (ctx->H[ix], (char *) p);
    }
}
